local optimalFps = 60
local lowFpsThreshold = 30
local highFpsThreshold = 80
local ultraHighFpsThreshold = 100  
local superUltraHighFpsThreshold = 120 
local currentLodDistanceBias = 0.3
local currentPedLodBias = 0.3
local currentVehicleLodBias = 0.3
local currentParticleLightMultiplier = 0.3
local currentPedDensity = 1.0
local currentVehicleDensity = 1.0
local targetLodDistanceBias = currentLodDistanceBias
local targetPedLodBias = currentPedLodBias
local targetVehicleLodBias = currentVehicleLodBias
local targetParticleLightMultiplier = currentParticleLightMultiplier
local targetPedDensity = currentPedDensity
local targetVehicleDensity = currentVehicleDensity
local previousFps = 0  
local previousAvgFps = 0  
local fpsStabilityThreshold = 5 
local activeNotifications = {}

function lerp(a, b, t)
    return a + (b - a) * t
end

function transitionGraphicsSettings(dt)
    local transitionSpeed = 0.1
    currentLodDistanceBias = lerp(currentLodDistanceBias, targetLodDistanceBias, transitionSpeed * dt)
    currentPedLodBias = lerp(currentPedLodBias, targetPedLodBias, transitionSpeed * dt)
    currentVehicleLodBias = lerp(currentVehicleLodBias, targetVehicleLodBias, transitionSpeed * dt)
    currentParticleLightMultiplier = lerp(currentParticleLightMultiplier, targetParticleLightMultiplier, transitionSpeed * dt)
    currentPedDensity = lerp(currentPedDensity, targetPedDensity, transitionSpeed * dt)
    currentVehicleDensity = lerp(currentVehicleDensity, targetVehicleDensity, transitionSpeed * dt)
    SetVisualSettingFloat("lodDistanceBias", currentLodDistanceBias)
    SetVisualSettingFloat("pedLodBias", currentPedLodBias)
    SetVisualSettingFloat("vehicleLodBias", currentVehicleLodBias)
    SetVisualSettingFloat("particleLightMultiplier", currentParticleLightMultiplier)
    SetPedDensityMultiplierThisFrame(currentPedDensity)
    SetVehicleDensityMultiplierThisFrame(currentVehicleDensity)
    SetRandomVehicleDensityMultiplierThisFrame(currentVehicleDensity)
end

function smartAiAdjustGraphics(fps)
    if fps < 20 then  
        targetLodDistanceBias = 0.1 
        targetPedLodBias = 0.1
        targetVehicleLodBias = 0.1
        targetParticleLightMultiplier = 0.1
        targetPedDensity = 0.0
        targetVehicleDensity = 0.0
        notifyPlayer("Very low FPS detected, setting graphics to ultra low and disabling NPCs")
    elseif previousFps > fps + 3 then  
        targetLodDistanceBias = math.max(targetLodDistanceBias - 0.05, 0.2)
        targetPedLodBias = math.max(targetPedLodBias - 0.05, 0.2)
        targetVehicleLodBias = math.max(targetVehicleLodBias - 0.05, 0.2)
        targetParticleLightMultiplier = math.max(targetParticleLightMultiplier - 0.05, 0.2)
        targetPedDensity = math.max(targetPedDensity - 0.05, 0.1)
        targetVehicleDensity = math.max(targetVehicleDensity - 0.05, 0.1)
        notifyPlayer("Significant FPS drop detected, lowering graphics settings and NPC density")
    elseif fps < lowFpsThreshold then
        targetLodDistanceBias = math.max(targetLodDistanceBias - 0.05, 0.2)
        targetPedLodBias = math.max(targetPedLodBias - 0.05, 0.2)
        targetVehicleLodBias = math.max(targetVehicleLodBias - 0.05, 0.2)
        targetParticleLightMultiplier = math.max(targetParticleLightMultiplier - 0.05, 0.2)
        targetPedDensity = math.max(targetPedDensity - 0.05, 0.2)
        targetVehicleDensity = math.max(targetVehicleDensity - 0.05, 0.2)
        notifyPlayer("Low FPS detected, lowering graphics and NPC settings")
    elseif fps > superUltraHighFpsThreshold then 
        targetLodDistanceBias = 1.0 
        targetPedLodBias = 1.0
        targetVehicleLodBias = 1.0
        targetParticleLightMultiplier = 1.0
        targetPedDensity = 1.0
        targetVehicleDensity = 1.0
        SetTimecycleModifier("cinematic")
        SetExtraTimecycleModifier("reflection_correct_ambient")
        notifyPlayer("Super ultra high FPS detected, setting graphics to super ultra high and enabling full NPC density")
    elseif fps > ultraHighFpsThreshold then
        targetLodDistanceBias = 0.6 
        targetPedLodBias = 0.6
        targetVehicleLodBias = 0.6
        targetParticleLightMultiplier = 0.6
        targetPedDensity = 0.6
        targetVehicleDensity = 0.6
        notifyPlayer("Ultra high FPS detected, setting graphics to ultra high and adjusting NPC density")
    elseif fps > highFpsThreshold then
        targetLodDistanceBias = math.min(targetLodDistanceBias + 0.05, 0.6)
        targetPedLodBias = math.min(targetPedLodBias + 0.05, 0.6)
        targetVehicleLodBias = math.min(targetVehicleLodBias + 0.05, 0.6)
        targetParticleLightMultiplier = math.min(targetParticleLightMultiplier + 0.05, 0.6)
        targetPedDensity = math.min(targetPedDensity + 0.05, 0.6)
        targetVehicleDensity = math.min(targetVehicleDensity + 0.05, 0.6)
        notifyPlayer("High FPS detected, increasing graphics and NPC density settings")
    else
        targetLodDistanceBias = 0.3
        targetPedLodBias = 0.3
        targetVehicleLodBias = 0.3
        targetParticleLightMultiplier = 0.3
        targetPedDensity = 0.3
        targetVehicleDensity = 0.3
        notifyPlayer("Stable FPS, adjusting graphics and NPC settings to standard")
    end

    if math.abs(previousAvgFps - fps) > fpsStabilityThreshold then
        targetLodDistanceBias = math.max(targetLodDistanceBias - 0.1, 0.2)
        targetPedLodBias = math.max(targetPedLodBias - 0.1, 0.2)
        targetVehicleLodBias = math.max(targetVehicleLodBias - 0.1, 0.2)
        targetParticleLightMultiplier = math.max(targetParticleLightMultiplier - 0.1, 0.2)
        targetPedDensity = math.max(targetPedDensity - 0.1, 0.2)
        targetVehicleDensity = math.max(targetVehicleDensity - 0.1, 0.2)
        notifyPlayer("Unstable FPS detected, adjusting graphics and NPC settings for stability")
    end
    
    previousFps = fps 
    previousAvgFps = fps  
end

function monitorFpsWithAi()
    local fps = 0
    local lastFpsValues = {}
    local maxFpsHistory = 10

    Citizen.CreateThread(function()
        while true do
            fps = 1 / GetFrameTime()
            table.insert(lastFpsValues, fps)
            if #lastFpsValues > maxFpsHistory then
                table.remove(lastFpsValues, 1)
            end
            local avgFps = 0
            for _, v in ipairs(lastFpsValues) do
                avgFps = avgFps + v
            end
            avgFps = avgFps / #lastFpsValues
            smartAiAdjustGraphics(avgFps)
            transitionGraphicsSettings(1)
            Citizen.Wait(10000)
        end
    end)
end

function disableMotionBlur()
    SetTimecycleModifier("default")
    SetExtraTimecycleModifier("default")
    SetVisualSettingFloat("motionblurStrength", 0.0)
    notifyPlayer("Motion blur disabled.")
end

function setOptimalGraphics()
    SetVisualSettingFloat("lodDistanceBias", 0.3)
    SetVisualSettingFloat("pedLodBias", 0.3)
    SetVisualSettingFloat("vehicleLodBias", 0.3)
    SetVisualSettingFloat("grassLodBias", 0.5)
    SetVisualSettingFloat("particleLightMultiplier", 0.3)
    SetTimecycleModifier("cinema")
    SetExtraTimecycleModifier("reflection_correct_ambient")
    disableMotionBlur()
    notifyPlayer("Optimal graphics settings applied.")
end

function notifyPlayer(message)
    if #activeNotifications >= 2 then
        RemoveNotification(table.remove(activeNotifications, 1))
    end
    local notification = BeginTextCommandThefeedPost("STRING")
    AddTextComponentSubstringPlayerName(message)
    EndTextCommandThefeedPostMessagetext("CHAR_LIFEINVADER", "CHAR_LIFEINVADER", false, false, "Graphics AI Notification", "")
    local notificationHandle = EndTextCommandThefeedPostTicker(true, false)
    table.insert(activeNotifications, notificationHandle)
    Citizen.SetTimeout(5000, function()
        for i, notif in ipairs(activeNotifications) do
            if notif == notificationHandle then
                RemoveNotification(notif)
                table.remove(activeNotifications, i)
                break
            end
        end
    end)
end

AddEventHandler('playerSpawned', function()
    Citizen.Wait(1000)
    setOptimalGraphics()
    monitorFpsWithAi()
end)

AddEventHandler('onClientMapStart', function()
    Citizen.Wait(1000)
    setOptimalGraphics()
    monitorFpsWithAi()
end)
